import warnings
import numpy as np
import scipy.stats as stats
import scipy.signal as signal
from pyyeti import dsp, locate, ytools
import pytest


def test_resample():
    p = 5
    q = 5
    n = 530
    np.random.seed(1)
    d = np.random.randn(n) - 55
    x = np.arange(n) + 100

    rtol = atol = 1e-15
    d0, x0, fir0 = dsp.resample(d, p, q, t=x, getfir=1)
    assert np.allclose(d0, d, rtol, atol)
    assert np.allclose(x0, x, rtol, atol)
    n = 2 * 10 + 1
    mid = (n - 1) // 2
    assert np.allclose(fir0[mid], 1, rtol, atol)
    fir0[mid] = 0
    assert np.allclose(0, fir0, rtol, atol)

    pts = 5
    d0, x0, fir0 = dsp.resample(d, p, q, t=x, getfir=1, pts=pts)
    assert np.allclose(d0, d, rtol, atol)
    assert np.allclose(x0, x, rtol, atol)
    n = 2 * pts + 1
    mid = (n - 1) // 2
    assert np.allclose(fir0[mid], 1, rtol, atol)
    fir0_ = fir0.copy()
    fir0[mid] = 0
    assert np.allclose(0, fir0, rtol, atol)
    d1, fir1 = dsp.resample(d, p, q, getfir=1, pts=pts)
    d2, x2 = dsp.resample(d, p, q, t=x, pts=pts)
    assert np.all(d1 == d0)
    assert np.all(d1 == d2)
    assert np.allclose(fir0_, fir1, rtol, atol)
    assert np.allclose(x2, x, rtol, atol)

    D = np.vstack((d, d)).T
    D2, X2 = dsp.resample(D, p, q, t=x, pts=pts, axis=0)
    assert np.all(d1 == D[:, 0])
    assert np.all(d1 == D[:, 1])
    assert np.allclose(x2, X2)

    D3, X3 = dsp.resample(D.T, p, q, t=x, pts=pts)
    assert np.allclose(D2, D3.T)
    assert np.allclose(X3, X2)

    lcc = {}
    for pts in [10, 50]:
        dup4, xup4 = dsp.resample(d, 4, 1, t=x, pts=pts)
        dup2, xup2 = dsp.resample(dup4, 10, 20, t=xup4, pts=pts)
        d0, x0 = dsp.resample(dup2, 1, 2, t=xup2, pts=pts)
        assert np.allclose(x0, x, rtol, atol)
        lcc[pts] = stats.pearsonr(d0, d)[0]
        assert 1 - lcc[pts] < 0.015
        assert np.abs(np.mean(d) - np.mean(d0)) < 0.005
    assert lcc[50] > lcc[10]


def test_exclusive_sgfilter():
    with pytest.raises(ValueError):
        dsp.exclusive_sgfilter([1, 2, 3, 4, 5, 6, 7, 8], 5, "badoption")
    with pytest.raises(ValueError):
        dsp.exclusive_sgfilter([1, 2, 3, 4, 5, 6, 7, 8], 5, 6)
    x = np.arange(4 * 2 * 150 * 3.0).reshape(4, 2, 150, 3)
    x1 = signal.savgol_filter(x, 11, 0, axis=2)
    x2 = dsp.exclusive_sgfilter(x, 11, exclude_point=None, axis=2)
    assert np.allclose(x1, x2)


def test_windowends():
    s = dsp.windowends(np.ones(8), 4)
    assert np.allclose([0.0, 0.25, 0.75, 1.0, 1.0, 1.0, 1.0, 1.0], s)
    s = dsp.windowends(np.ones(8), 0.7, ends="back")
    assert np.allclose(
        [1.0, 1.0, 1.0, 1.0, 0.85355339059327373, 0.5, 0.14644660940672621, 0.0], s
    )
    s = dsp.windowends(np.ones(8), 0.5, ends="both")
    assert np.allclose([0.0, 0.25, 0.75, 1.0, 1.0, 0.75, 0.25, 0.0], s)
    s = dsp.windowends(np.ones(8), 0.5, ends="none")
    assert np.allclose([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0], s)
    s = dsp.windowends(np.ones(8), 3)
    assert np.allclose([0.0, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0], s)
    s = dsp.windowends(np.ones(8), 2)
    assert np.allclose([0.0, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0], s)


def test_waterfall():
    from pyyeti import srs

    sig, t, f = ytools.gensweep(10, 1, 50, 4)
    sr = 1 / t[1]
    frq = np.arange(1.0, 5.1)
    Q = 20
    sig2 = sig[: int(sr * 1.5), None]

    def func(s):
        return srs.srs(s, sr, frq, Q), frq

    mp, t, f = dsp.waterfall(sig, sr, 2, 0.5, func, which=0, freq=1)
    mp, t, f = dsp.waterfall(sig2, sr, 2, 0.5, func, which=0, freq=1)

    def func2(s):
        return srs.srs(s, sr, frq, Q)

    mp2, t2, f2 = dsp.waterfall(sig2, sr, 2, 0.5, func2, which=None, freq=frq)

    assert np.allclose(mp, mp2)
    assert np.allclose(t, t2)
    assert np.allclose(f, f2)

    with pytest.raises(ValueError):
        dsp.waterfall(sig, sr, 2, 0.5, func, which=None, freq=1)
    with pytest.raises(ValueError):
        dsp.waterfall(sig, sr, 2, 1.5, func, which=None, freq=frq)
    with pytest.raises(ValueError):
        dsp.waterfall(sig, sr, 2, 1.0, func, which=None, freq=frq)
    with pytest.raises(ValueError):
        dsp.waterfall(sig, sr, 2, -0.5, func, which=None, freq=frq)
    sig = np.hstack((sig2, sig2))
    with pytest.raises(ValueError):
        dsp.waterfall(sig, sr, 2, 0.5, func, which=None, freq=frq)


def test_waterfall2():
    N = 5000
    sig = np.ones(N)
    _t = np.arange(N) / 1000.0
    sr = 1 / _t[1]
    frq = np.arange(1.0, 5.1)

    def func(s):
        return np.ones(len(frq)), frq

    # test different overlaps
    S, Si = 1.0, "1000"
    O, Oi = 0.5, "500"
    mp, t, f = dsp.waterfall(sig, sr, S, O, func, which=0, freq=1)
    mpi, ti, fi = dsp.waterfall(sig, sr, Si, Oi, func, which=0, freq=1)
    assert np.allclose(mp, mpi)
    assert np.allclose(t, ti)
    assert np.allclose(f, fi)
    step = S * (1 - O)
    tcmp = np.arange(S / 2.0, _t[-1] - S / 2 + step / 2, step)
    assert np.allclose(t, tcmp)

    O, Oi = 0.9, "900"
    mp, t, f = dsp.waterfall(sig, sr, S, O, func, which=0, freq=1)
    mpi, ti, fi = dsp.waterfall(sig, sr, Si, Oi, func, which=0, freq=1)
    assert np.allclose(mp, mpi)
    assert np.allclose(t, ti)
    assert np.allclose(f, fi)
    step = S * (1 - O)
    tcmp = np.arange(S / 2.0, _t[-1] - S / 2 + step / 2, step)
    assert np.allclose(t, tcmp)

    O, Oi = 0.1, "100"
    mp, t, f = dsp.waterfall(sig, sr, S, O, func, which=0, freq=1)
    mpi, ti, fi = dsp.waterfall(sig, sr, Si, Oi, func, which=0, freq=1)
    assert np.allclose(mp, mpi)
    assert np.allclose(t, ti)
    assert np.allclose(f, fi)
    step = S * (1 - O)
    tcmp = np.arange(S / 2.0, _t[-1] - S / 2 + step / 2, step)
    assert np.allclose(t, tcmp)

    with pytest.raises(ValueError):
        dsp.waterfall(sig, sr, Si, -1, func, which=0, freq=1)
    with pytest.raises(ValueError):
        dsp.waterfall(sig, sr, Si, Si, func, which=0, freq=1)


def test_get_turning_pts():
    y, x = dsp.get_turning_pts([1, 2, 3, 3, 3], [1, 2, 3, 4, 5], getindex=False)
    assert np.all([1, 3, 3] == y)
    assert np.all([1, 3, 5] == x)
    y2 = dsp.get_turning_pts([1, 2, 3, 3, 3], getindex=False)
    assert np.all(y == y2)
    with pytest.raises(ValueError):
        dsp.get_turning_pts([1, 2, 3, 3, 3], [1, 2, 3, 3, 5])
    with pytest.raises(ValueError):
        dsp.get_turning_pts([1, 2, 3, 3, 3, 5], [1, 2, 3, 4, 5])


def test_calcenv():
    p = 5
    x = [1, 2, 3, 2, 5]
    y = [1, 4, 9, 4, 2]
    with pytest.raises(ValueError):
        dsp.calcenv(x, y, p)

    x = [1, 2, 3, 4, 5]
    y = [1, 4, 9, 16]
    with pytest.raises(ValueError):
        dsp.calcenv(x, y, p)

    with pytest.raises(ValueError):
        dsp.calcenv([0, 1, 2], [0, 5, 0], method="default")

    xe, ye, *_ = dsp.calcenv([0, 1, 2], [0, 5, 0], makeplot="no")
    x_sb = [0, 0.95, 1.05, 2 * 1.05]
    y_sb = [0, 5, 5, 0]
    yi = np.interp(xe, x_sb, y_sb)
    err = abs(yi - ye).max()
    assert err < 0.006

    xen, yen, *_ = dsp.calcenv([0, 1, 2], [0, -5, 0], makeplot="no", method="min")
    assert np.allclose(xe, xen)
    assert np.allclose(ye, -yen)

    xe, ye, *_ = dsp.calcenv([0, 1, 2], [0, 5, 0], p=3, n=4000, makeplot="no")
    x_sb = [0, 0.97, 1.03, 2 * 1.03]
    y_sb = [0, 5, 5, 0]
    yi = np.interp(xe, x_sb, y_sb)
    assert abs(yi - ye).max() < err / 2

    xe, ye, *_ = dsp.calcenv([0, 1, 2], [0, -5, 0], p=20, makeplot="no", method="max")
    # find where the 2 line segments meet
    # first, get the line segments:
    #  y1 => (0, 0) to (1.20*1, -5): y1 = -5/1.2 x
    #  y2 => (.8, -5) to (.8*2, 0):  y2 = 5/.8 x + b
    #     get b:  0 = 5/.8 * 1.6 + b; b = -10
    #  y1 = -5/1.2 x
    #  y2 = 6.25 x - 10
    #  --> intersect @ x = 10 / (6.25+5/1.2)
    x_int = 10 / (6.25 + 5 / 1.2)
    y_int = -5 / 1.2 * x_int
    x_sb = [0, x_int, 1.6]
    y_sb = [0, y_int, 0]
    yi = np.interp(xe, x_sb, y_sb)
    assert abs(yi - ye).max() < 0.006

    xen, yen, *_ = dsp.calcenv([0, 1, 2], [0, 5, 0], p=20, makeplot="no", method="min")
    assert np.allclose(xe, xen)
    assert np.allclose(ye, -yen)

    (xmax, ymax, xmin, ymin, h) = dsp.calcenv(
        [0, 1, 2], [0, -5, 0], n=200, p=20, makeplot="no", base=None
    )
    yi = np.interp(xmax, x_sb, y_sb)
    assert abs(yi - ymax).max() < 0.06
    x_sb = [0, 0.8, 1.2, 2 * 1.2]
    y_sb = [0, -5, -5, 0]
    yi = np.interp(xmin, x_sb, y_sb)
    assert abs(yi - ymin).max() < 0.06

    (xmax, ymax, xmin, ymin, h) = dsp.calcenv(
        [0, 1, 2], [0, -5, 0], n=200, p=20, makeplot="no", method="both"
    )
    assert np.all(xmax == [0, 2])
    assert np.all(ymax == [0, 0])
    assert h is None
    x_sb = [0, 0.8, 1.2, 2 * 1.2]
    y_sb = [0, -5, -5, 0]
    yi = np.interp(xmin, x_sb, y_sb)
    assert abs(yi - ymin).max() < 0.06

    # test base:
    base = 2
    xe, ye, *_ = dsp.calcenv(
        [0, 1, 2], [0 - base, 5 - base, 0 - base], makeplot="no", n=100
    )
    xeb, yeb, *_ = dsp.calcenv([0, 1, 2], [0, 5, 0], makeplot="no", base=base, n=100)
    assert np.allclose(xe, xeb)
    assert np.allclose(ye + base, yeb)

    base = 2
    xe, ye, *_ = dsp.calcenv(
        [0, 1, 2], [0 - base, 5 - base, 0 - base], makeplot="no", n=100, method="min"
    )
    xeb, yeb, *_ = dsp.calcenv(
        [0, 1, 2], [0, 5, 0], makeplot="no", base=base, n=100, method="min"
    )
    assert np.allclose(xe, xeb)
    assert np.allclose(ye + base, yeb)

    base = 2
    (xex, yex, xen, yen, h) = dsp.calcenv(
        [0, 1, 2], [0 - base, 5 - base, 0 - base], makeplot="no", n=100, method="both"
    )
    (xexb, yexb, xenb, yenb, h) = dsp.calcenv(
        [0, 1, 2], [0, 5, 0], makeplot="no", base=base, n=100, method="both"
    )
    assert np.allclose(xex, xexb)
    assert np.allclose(yex + base, yexb)
    assert np.allclose(xen, xenb)
    assert np.allclose(yen + base, yenb)


def test_find_closest_times():
    a = np.array([1, 2, 3])
    b = np.array([0, 2.48, 2.51, 4])
    i = dsp._find_closest_times(a, b)
    assert np.all(i == [0, 1, 2, 2])

    v = np.arange(100.0)
    v2 = np.arange(0.0, 76.0, 0.1)
    j = dsp._find_closest_times(v, v2)
    assert 0.1 < (v2 - v[j]).max() < 0.51
    assert -0.51 < (v2 - v[j]).min() < -0.1


def test_find_closest_previous_times():
    a = np.array([1, 2, 3])
    b = np.array([0, 2.48, 2.51, 4])
    i = dsp._find_closest_previous_times(a, b)
    assert np.all(i == [0, 1, 1, 2])

    v = np.arange(100.0)
    v2 = np.arange(0.0, 76.0, 0.1)
    j = dsp._find_closest_previous_times(v, v2)
    assert 0.51 < (v2 - v[j]).max() < 1.01
    assert (v2 - v[j]).min() >= 0.0


def test_fixtime():
    with pytest.raises(ValueError):
        dsp.fixtime(([1, 2, 3, 4], [1, 2, 3]))

    t = [0, 1, 6, 7]
    y = [1, 2, 3, 4]
    with pytest.warns(RuntimeWarning):
        tn, yn = dsp.fixtime((t, y), sr=1, verbose=False)
        ty2 = dsp.fixtime(np.vstack([t, y]).T, sr=1, verbose=False)

    t2, y2 = ty2.T
    assert np.all(tn == t2)
    assert np.all(yn == y2)
    assert np.all(tn == np.arange(8))
    assert np.all(yn == [1, 2, 2, 2, 3, 3, 3, 4])

    with pytest.warns(RuntimeWarning):
        (t2, y2), info = dsp.fixtime((t, y), getall=True, sr=1, verbose=False)
    assert np.all(tn == t2)
    assert np.all(yn == y2)
    assert np.all(info.alldrops.dropouts == [])
    assert np.allclose(info.sr_stats, [1, 0.2, 3 / 7, 1, 200 / 3])
    assert np.all(info.tp == np.arange(4))
    assert info.despike_info is None
    with pytest.raises(ValueError):
        dsp.fixtime((1, 1, 1))

    drop = -1.40130e-45
    t = [0, 0.5, 1, 6, 7]
    y = [1, drop, 2, 3, 4]

    with pytest.warns(RuntimeWarning):
        t2, y2 = dsp.fixtime([t, y], sr=1, verbose=False)
    assert np.all(tn == t2)
    assert np.all(yn == y2)

    t = [1, 2, 3]
    y = [drop, drop, drop]
    pytest.warns(RuntimeWarning, dsp.fixtime, (t, y), "auto")

    (t2, y2), info = dsp.fixtime(
        (t, y), "auto", deldrops=False, getall=True, verbose=False
    )
    assert np.all(info.tp == [0, 2])
    assert np.all(t2 == t)
    assert np.all(y2 == y)
    assert info.alldrops.dropouts is None

    t = np.arange(100.0)
    t = np.hstack((t, 200.0))
    y = np.arange(101)
    with pytest.warns(RuntimeWarning):
        t2, y2 = dsp.fixtime((t, y), "auto", verbose=False)
    assert np.all(t2 == np.arange(100.0))
    assert np.all(y2 == np.arange(100))

    with pytest.warns(RuntimeWarning):
        t2, y2 = dsp.fixtime(
            (t, y), "auto", delouttimes=0, hold_previous_value=True, verbose=False
        )
    assert np.all(t2 == np.arange(200.1))
    sbe = np.ones(201, int) * 99
    sbe[:100] = np.arange(100)
    sbe[-1] = 100
    assert np.all(y2 == sbe)

    with pytest.warns(RuntimeWarning):
        t2, y2 = dsp.fixtime((t, y), "auto", delouttimes=0, verbose=False)
    assert np.all(t2 == np.arange(200.1))
    sbe = np.ones(201, int) * 99
    sbe[:100] = np.arange(100)
    sbe[150:] = 100
    assert np.all(y2 == sbe)

    t = [0, 1, 2, 3, 4, 5, 7, 6]
    y = [0, 1, 2, 3, 4, 5, 7, 6]
    with pytest.warns(RuntimeWarning):
        t2, y2 = dsp.fixtime((t, y), 1, verbose=False)
    assert np.all(t2 == np.arange(8))
    assert np.all(y2 == np.arange(8))

    with pytest.raises(ValueError):
        dsp.fixtime((t, y), 1, negmethod="stop", verbose=False)

    t = np.arange(8)[::-1]
    with pytest.raises(ValueError):
        dsp.fixtime((t, y), 1, verbose=False)

    t = np.arange(0, 1, 0.001)
    noise = np.random.randn(t.size)
    noise = noise / noise.max() / 100000
    y = np.random.randn(t.size)
    t2, y2 = dsp.fixtime((t + noise, y), "auto", base=0, verbose=False)
    assert np.allclose(t, t2)
    assert np.all(y == y2)

    dt = np.ones(100)
    dt[80:] = 0.91 * dt[80:]
    t = dt.cumsum()
    y = np.arange(100)
    with pytest.warns(RuntimeWarning):
        t2, y2 = dsp.fixtime((t, y), 1, verbose=False)

    dt = np.ones(100)
    dt[80:] = 1.08 * dt[80:]
    t = dt.cumsum()
    y = np.arange(100)
    with pytest.warns(RuntimeWarning):
        t2, y2 = dsp.fixtime((t, y), 1, verbose=False)


def test_fixtime_prev_val_tol():
    t = [0, 1, 6.01, 7]
    y = [1, 2, 3, 4]
    with pytest.warns(RuntimeWarning):
        yn = dsp.fixtime((t, y), sr=1, hold_previous_value=True, verbose=False)[1]

    assert np.all(yn == [1, 2, 2, 2, 2, 2, 2, 4])

    with pytest.warns(RuntimeWarning):
        yn = dsp.fixtime(
            (t, y),
            sr=1,
            hold_previous_value=True,
            verbose=False,
            previous_value_tol=0.1,
        )[1]

    assert np.all(yn == [1, 2, 2, 2, 2, 2, 3, 4])

    t = [0, 1, 2, 3]
    y = [1, 2, 3, 4]
    with pytest.raises(ValueError):
        t2, y2 = dsp.fixtime(
            (t, y),
            sr=1,
            hold_previous_value=True,
            previous_value_tol=1.5,
            verbose=False,
        )

    with pytest.raises(ValueError):
        t2, y2 = dsp.fixtime(
            (t, y),
            sr=1,
            hold_previous_value=True,
            previous_value_tol=-0.2,
            verbose=False,
        )


def test_fixtime2():
    dt = 0.001
    t = np.arange(0, 1, dt)
    t[1:] += dt / 2
    # t[400:] -= dt/2
    noise = np.random.randn(t.size)
    noise = noise / noise.max() / 100000
    y = np.random.randn(t.size)

    t2, y2 = dsp.fixtime((t + noise, y), "auto", base=None, verbose=False)
    pv = locate.find_subseq(y2, y[1:-1])
    assert pv.size > 0

    t3, y3 = dsp.fixtime((t + noise, y), "auto", base=0.0, verbose=False)
    pv = locate.find_subseq(y2, y[1:-1])
    assert pv.size > 0
    assert not np.allclose(t2[0], t3[0])
    assert np.allclose(t2 - t2[0], t3 - t3[0])
    assert abs(t3 - 0.003).min() < 1e-14


def test_fixtime_too_many_tp():
    # actual data:
    fd = np.array(
        [
            [0.91878125, 1.17070317],
            [0.9226875, 1.17070317],
            [0.9226875, 1.17070317],
            [0.92659375, 1.17070317],
            [0.92659375, 1.17070317],
            [0.9305, 1.17070317],
            [0.9305, 1.17070317],
            [0.93440625, 1.17070317],
            [0.93440625, 1.17070317],
            [0.9383125, 1.17070317],
            [0.9383125, 1.36582029],
            [0.94221875, 1.29265141],
            [0.94221875, 0.97558594],
            [0.946125, 1.63410652],
            [0.946125, 0.92680663],
            [0.95003125, 1.5365479],
            [0.95003125, 1.07314456],
            [0.9539375, 1.26826179],
            [0.9539375, 1.31704104],
            [0.95784375, 1.1950928],
            [0.95784375, 1.29265141],
            [0.96175, 1.04875493],
            [0.96175, 1.34143066],
            [0.96565625, 0.92680663],
            [0.96565625, 1.36582029],
            [0.9695625, 0.87802738],
            [0.9695625, 1.24387205],
            [0.97346875, 0.97558594],
            [0.97346875, 1.14631355],
            [0.977375, 0.99997562],
            [0.977375, 1.14631355],
            [0.98128125, 1.07314456],
            [0.98128125, 1.07314456],
            [0.9851875, 1.1950928],
            [0.9851875, 1.09753418],
            [0.98909375, 1.1950928],
            [0.98909375, 1.1219238],
            [0.993, 1.1950928],
            [0.993, 1.17070317],
            [0.99690625, 1.1950928],
            [0.99690625, 1.14631355],
            [1.0008125, 1.17070317],
        ]
    )

    with pytest.raises(ValueError):
        dsp.fixtime(([1, 2, 3, 4], [1, 2, 3]), verbose=False)
    with pytest.raises(ValueError):
        dsp.fixtime(np.random.randn(3, 3, 3), verbose=False)

    with pytest.warns(RuntimeWarning):
        t, d = dsp.fixtime((*fd.T,), sr="auto", verbose=False)
    # assert np.all(fd[:, 1] == d)
    assert np.allclose(np.diff(t), 0.002)
    assert np.allclose(t[0], fd[0, 0])


def test_fixtime_naninf():
    t = np.arange(15)
    y = [1, 2, 3, 4, np.nan, 6, 7, np.inf, 9, 10, -np.inf, 12, -1.40130e-45, 14, 15]

    with pytest.warns(RuntimeWarning):
        (tn, yn), info = dsp.fixtime((t, y), sr=1, getall=1, verbose=False)

    d = info.alldrops.dropouts
    assert np.all(d == [4, 7, 10, 12])
    assert np.all(t == tn)
    assert np.all(yn == [1, 2, 3, 4, 4, 6, 7, 7, 9, 10, 10, 12, 12, 14, 15])


def test_fixtime_despike():
    t = np.arange(16)
    y = [
        1,
        2,
        3,
        4,
        np.nan,
        6,
        7,
        np.inf,
        9,
        10,
        -np.inf,
        12,
        -1.40130e-45,
        14,
        134,
        15,
    ]

    with pytest.warns(RuntimeWarning):
        (tn, yn), info = dsp.fixtime(
            (t, y),
            sr=1,
            getall=1,
            hold_previous_value=True,
            delspikes=dict(n=9, sigma=2),
            verbose=False,
        )
        (tn2, yn2), info2 = dsp.fixtime(
            (t, y),
            sr=1,
            getall=1,
            hold_previous_value=True,
            delspikes=dict(sigma=2),
            verbose=False,
        )

    d = info.alldrops.dropouts
    assert np.all(d == range(4, 13))
    assert np.all(t == tn)
    assert np.all(t == tn2)
    assert np.allclose(yn, [1, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 15])
    assert np.allclose(yn2, yn)
    assert info.despike_info.delspikes["n"] == 9
    assert info2.despike_info.delspikes["n"] == 15
    # spike is at pos 14
    assert np.all(info.alldrops.spikes == 14)
    assert np.all(info2.alldrops.spikes == 14)
    assert np.all(info.alldrops.alldrops == range(4, 15))


def test_fixtime_perfect():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.random.randn(len(t))
    t2, y2 = dsp.fixtime((t, y), "auto", verbose=False)
    assert np.allclose(t2, t)
    assert np.all(y2 == y)


def test_fixtime_perfect_delspike():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.sin(2 * np.pi * 5 * t)
    (t2, y2), info = dsp.fixtime(
        (t, y), "auto", delspikes=True, getall=1, verbose=False
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == y)

    with pytest.raises(ValueError):
        dsp.fixtime(
            (t, y),
            "auto",
            delspikes=dict(method="badmethod"),
            getall=1,
            verbose=False,
        )


def test_fixtime_exclude_point_change():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.empty(1000)
    y[::2] = 0.1
    y[1::2] = -0.1
    ycopy = y.copy()
    ycopy[100:105] = y[99]
    y[100:105] = 10.0

    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=True,
        verbose=False,
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)

    t2, y2 = dsp.fixtime(
        (t, y), "auto", delspikes=dict(exclude_point="last"), verbose=False
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)

    with pytest.raises(ValueError):
        dsp.fixtime(
            (t, y),
            "auto",
            delspikes=dict(exclude_point="middle"),
            verbose=False,
        )

    # some of a ramp up should be deleted:
    y[100:105] = np.arange(0.0, 5.0) * 3
    ycopy = y.copy()
    ycopy[102:105] = y[101]
    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=dict(sigma=18),
        verbose=False,
        hold_previous_value=True,
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)

    # none of a ramp down should be deleted:
    y[100:105] = np.arange(5.0, 0.0, -1.0) * 3
    t2, y2 = dsp.fixtime((t, y), "auto", delspikes=dict(sigma=18), verbose=False)
    assert np.allclose(t2, t)
    assert np.all(y2 == y)


def test_fixtime_exclude_point_change2():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.empty(1000)
    y[::2] = 0.1
    y[1::2] = -0.1
    ycopy = y.copy()
    ycopy[100:105] = y[99]
    y[100:105] = 10.0

    t2, y2 = dsp.fixtime(
        (t, y), "auto", delspikes=dict(method="despike"), verbose=False
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)

    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=dict(exclude_point="last", method="despike"),
        verbose=False,
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)

    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=dict(exclude_point="middle", method="despike"),
        verbose=False,
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == y)

    # some of a ramp up should be deleted:
    y[100:105] = np.arange(0.0, 5.0) * 3
    ycopy = y.copy()
    ycopy[102:105] = y[101]
    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=dict(sigma=35, method="despike"),
        verbose=False,
        hold_previous_value=True,
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)

    # none of a ramp down should be deleted:
    y[100:105] = np.arange(5.0, 0.0, -1.0) * 3
    t2, y2 = dsp.fixtime(
        (t, y), "auto", delspikes=dict(sigma=35, method="despike"), verbose=False
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == y)


def test_fixtime_simple_despike():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.empty(1000)
    y[::2] = 0.1
    y[1::2] = -0.1
    ycopy = y.copy()
    ycopy[100] = y[99]
    y[100] = 10.0

    t2, y2 = dsp.fixtime(
        (t, y), "auto", delspikes=dict(method="simple", sigma=4), verbose=False
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)

    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=dict(method="simple", sigma=4, maxiter=1),
        verbose=False,
    )
    assert np.allclose(t2, t)
    assert np.all(y2 == ycopy)


def test_fixtime_despike_edge_conditions():
    dt = 0.001
    n = 45
    t = np.arange(n) * dt
    y = np.empty(n)
    for i in range(1, n - 5):
        y[::2] = 0.1
        y[1::2] = -0.1
        ycopy = y.copy()
        s = slice(i, i + 5)
        ycopy[s] = y[i - 1]
        y[s] = 10.0

        # with despike_diff:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            t2, y2 = dsp.fixtime((t, y), "auto", delspikes=True, verbose=False)

        assert np.allclose(t2, t)
        if i < n - 15:
            assert np.allclose(y2, ycopy)
        else:
            assert np.allclose(y2, y)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            t2, y2 = dsp.fixtime(
                (t, y), "auto", delspikes=dict(exclude_point="last"), verbose=False
            )

        assert np.allclose(t2, t)
        if i > 10:
            assert np.allclose(y2, ycopy)
        else:
            assert np.allclose(y2, y)

        # now with despike:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            t2, y2 = dsp.fixtime(
                (t, y), "auto", delspikes=dict(method="despike"), verbose=False
            )
        assert np.allclose(t2, t)
        if i < n - 15 - 3:
            assert np.allclose(y2, ycopy)
        else:
            assert np.allclose(y2, y)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            t2, y2 = dsp.fixtime(
                (t, y),
                "auto",
                delspikes=dict(method="despike", exclude_point="last"),
                verbose=False,
            )
        assert np.allclose(t2, t)
        if i > 13:
            assert np.allclose(y2, ycopy)
        else:
            assert np.allclose(y2, y)


def test_fix_edges2():
    dt = 0.01
    t = np.arange(50) * dt

    y = np.zeros(50)
    n = 44
    y[n:] = 1
    t2, y2 = dsp.fixtime(
        (t, y), "auto", delspikes=dict(exclude_point="last"), verbose=False
    )
    assert np.allclose(t2, t[:n])
    assert np.all(y2 == y[:n])

    y = np.zeros(50)
    n = 5
    y[:n] = 1
    t2, y2 = dsp.fixtime(
        (t, y), "auto", delspikes=dict(exclude_point="first"), verbose=False
    )
    assert np.allclose(t2, t[n:])
    assert np.all(y2 == y[n:])

    y = np.zeros(50)
    n = 44
    y[n:] = 1
    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=dict(exclude_point="last", method="despike"),
        verbose=False,
    )
    assert np.allclose(t2, t[:n])
    assert np.all(y2 == y[:n])

    y = np.zeros(50)
    n = 5
    y[:n] = 1
    t2, y2 = dsp.fixtime(
        (t, y),
        "auto",
        delspikes=dict(exclude_point="first", method="despike"),
        verbose=False,
    )
    assert np.allclose(t2, t[n:])
    assert np.all(y2 == y[n:])


def test_fixtime_wrong_sr():
    t = np.arange(10)
    y = t + 10
    msg = ".*10 vs 9.*sample rate used too low[?] Only aligning on.*"
    with pytest.warns(RuntimeWarning, match=msg):
        (t2, y2), info = dsp.fixtime((t, y), getall=True, sr=0.94, verbose=False)

    assert t2[0] == 0.0
    assert t2[-1] < 9.0

    msg = ".*10 vs 11.*sample rate used too high[?] Only aligning on.*"
    with pytest.warns(RuntimeWarning, match=msg):
        (t2, y2), info = dsp.fixtime((t, y), getall=True, sr=1.06, verbose=False)

    assert t2[0] == 0.0
    assert t2[-1] > 9.0


def test_aligntime():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.random.randn(len(t))

    t2 = np.arange(100, 800) * dt
    y2 = np.random.randn(len(t2))

    t3 = np.arange(200, 850) * dt
    y3 = np.random.randn(len(t3))

    dct = dict(ax1=(t, y), ax2=(t2, y2), ax3=(t3, y3))
    newdct = dsp.aligntime(dct)
    assert abs(newdct["t"][0] - t3[0]) < 1e-14
    print(newdct["t"][-1], t2[-1], abs(newdct["t"][-1] - t2[-1]))
    assert abs(newdct["t"][-1] - t2[-1]) < 1e-14

    newdct = dsp.aligntime(dct, ["ax1", "ax3"])
    assert abs(newdct["t"][0] - t3[0]) < 1e-14
    assert abs(newdct["t"][-1] - t3[-1]) < 1e-14

    with pytest.raises(ValueError):
        dsp.aligntime(dct, ["ax1", "ax4"])

    newdct = dsp.aligntime(dct, mode="expand")
    assert abs(newdct["t"][0] - t[0]) < 1e-14
    assert abs(newdct["t"][-1] - t[-1]) < 1e-14

    t4 = np.arange(1200, 1350) * dt
    y4 = np.random.randn(len(t4))
    dct["ax4"] = t4, y4
    with pytest.raises(ValueError):
        dsp.aligntime(dct, ["ax1", "ax4"])


def test_aligntime2():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.random.randn(len(t))

    dt = 0.0011
    t2 = np.arange(100, 800) * dt
    y2 = np.random.randn(len(t2))

    dct = dict(ax1=(t, y), ax2=(t2, y2))
    with pytest.raises(ValueError):
        dsp.aligntime(dct)
    with pytest.raises(ValueError):
        dsp.aligntime(dct, mode="expand")


def test_fdscale():
    sig, t, f = ytools.gensweep(10, 1, 12, 8)
    scale = np.array(
        [[0.0, 1.0], [4.0, 1.0], [5.0, 0.5], [8.0, 0.5], [9.0, 1.0], [100.0, 1.0]]
    )
    sr = 1 / t[1]
    sig_scaled = dsp.fdscale(sig, sr, scale)
    x = np.arange(0, 10, 0.001)
    y = 1 / np.interp(x, scale[:, 0], scale[:, 1])
    unscale = np.vstack((x, y)).T
    sig_unscaled = dsp.fdscale(sig_scaled, sr, unscale)
    assert np.allclose(sig, sig_unscaled, atol=1e-6)

    sig_scaled = dsp.fdscale(sig[:-1], sr, scale)
    sig_unscaled = dsp.fdscale(sig_scaled, sr, unscale)
    assert np.allclose(sig[:-1], sig_unscaled, atol=1e-6)

    sig2 = np.vstack((sig, sig)).T
    sig_scaled = dsp.fdscale(sig2, sr, scale)
    sig_unscaled = dsp.fdscale(sig_scaled, sr, unscale)
    assert np.allclose(sig2, sig_unscaled, atol=1e-6)

    sign = np.vstack((sig, sig, sig))
    sigf1 = dsp.fdscale(sign, sr, scale)
    sigf2 = dsp.fdscale(sign.T, sr, scale, axis=0).T
    assert np.allclose(sigf1, sigf2)

    sign3 = sign.reshape(1, 3, -1, 1)
    sigf3 = dsp.fdscale(sign3, sr, scale, axis=2).reshape(3, -1)
    assert np.allclose(sigf1, sigf3)


def test_fftfilt():
    # make a signal of sinusoids for testing:
    h = 0.001
    t = np.arange(0, 3.0 + h / 2, h)
    y1 = 10 + 3.1 * np.sin(2 * np.pi * 3 * t)
    y2 = 5 * np.sin(2 * np.pi * 10 * t)
    y3 = 2 * np.sin(2 * np.pi * 30 * t)
    y4 = 3 * np.sin(2 * np.pi * 60 * t)
    y = y1 + y2 + y3 + y4

    sr = 1 / h
    nyq = sr / 2
    for j, (w, pz, yj) in enumerate(
        ((7, None, y1), ([7, 18], None, y2), ([18, 45], None, y3), (45, False, y4))
    ):
        yf = dsp.fftfilt(y, w, bw=None, pass_zero=pz, nyq=nyq)[0]
        assert 1 - stats.pearsonr(yf, yj)[0] < 0.01

    yf = dsp.fftfilt(y, [7, 45], pass_zero=True, nyq=nyq)[0]
    assert 1 - stats.pearsonr(yf, y1 + y4)[0] < 0.01

    yf = dsp.fftfilt(y, [7, 18, 45], nyq=nyq)[0]
    assert 1 - stats.pearsonr(yf, y1 + y3)[0] < 0.01
    assert yf.ndim == 1

    with pytest.raises(ValueError):
        dsp.fftfilt(y, 2)
    with pytest.raises(ValueError):
        dsp.fftfilt(y, [0.1, 0.3], bw=[0.2, 0.2, 0.2])

    y2 = np.column_stack((y, y))
    yf2 = dsp.fftfilt(y2, np.array([7, 18, 45]) / nyq, axis=0)[0]
    # yf2 = dsp.fftfilt(y2, np.array([7, 18, 45]), nyq=nyq)[0]
    assert np.allclose(yf2[:, 0], yf2[:, 1])
    assert np.allclose(yf2[:, 0], yf)
    assert yf2.shape[0] == y1.shape[0]

    yf, freq, H = dsp.fftfilt(y, [7, 18, 45], nyq=nyq)
    assert freq.shape == H.shape
    assert freq[0] == 0.0
    assert nyq - freq[-1] < freq[1]

    n2 = dsp.nextpow2(len(t))
    period = n2 * h
    assert np.allclose(freq[1], 1 / period)

    sig = y
    sign = np.vstack((sig, sig, sig))
    sigf1, *_ = dsp.fftfilt(sign, [7, 18, 45], nyq=nyq)
    sigf2, *_ = dsp.fftfilt(sign.T, [7, 18, 45], nyq=nyq, axis=0)
    assert np.allclose(sigf1, sigf2.T)

    sign3 = sign.reshape(1, 3, -1, 1)
    sigf3, *_ = dsp.fftfilt(sign3, [7, 18, 45], nyq=nyq, axis=2)
    assert np.allclose(sigf1, sigf3.reshape(3, -1))


def test_fftfilt2():
    s = np.random.randn(10000)
    sr = 4100.0
    with pytest.raises(ValueError):
        dsp.fftfilt(s, 2, nyq=sr / 2)
    with pytest.raises(ValueError):
        dsp.fftfilt(s, [300, 310], nyq=sr / 2)
    with pytest.raises(ValueError):
        dsp.fftfilt(s, [300, 4100 / 2 - 2], nyq=sr / 2)


def test_despike1():
    x = np.arange(100.0)
    x[45] = 4.5
    x[0] = -20
    x[-1] = 110
    s = dsp.despike(x, 9, sigma=4)
    pv = np.zeros(x.shape, bool)
    pv[0] = True
    pv[45] = True
    pv[-1] = True
    assert (s.pv == pv).all()
    assert (s.x == x[~pv]).all()
    with pytest.raises(ValueError):
        dsp.despike(np.ones((5, 5)), 3)


def test_despike2():
    dt = 0.001
    t = np.arange(1000) * dt
    y = np.sin(2 * np.pi * 5 * t)
    y[400:500] = 0.0
    y[450] = 0.3
    y1 = np.concatenate((y[:450], y[451:]))
    pv = np.zeros(1000, bool)
    pv[450] = True
    desp = dsp.despike(y, 15)
    assert np.all(desp.x == y1)
    assert np.all(desp.pv == pv)

    desp = dsp.despike(y, 15, threshold_value=0.29)
    assert np.all(desp.x == y1)
    assert np.all(desp.pv == pv)

    desp = dsp.despike(y, 15, threshold_value=0.31)
    assert np.all(desp.x == y)
    pv = np.zeros(1000, bool)
    assert np.all(desp.pv == pv)

    y[450] = 0.05
    desp = dsp.despike(y, 15, threshold_value=0.051)
    assert np.all(desp.x == y)
    assert np.all(desp.pv == pv)


def test_despike3():
    # tests the sweep_out_priors feature:
    a = np.ones(1500)
    a[75:100] = 5.0
    a[115:125] = 5.0
    a[575:600] = 5.0
    desp = dsp.despike(a, 35)
    assert desp.niter < 5
    pv = np.zeros(1500)
    pv[75:100] = True
    pv[115:125] = True
    pv[575:600] = True
    assert (desp.pv == pv).all()


def test_despike4():
    # tests the sweep_out_nexts feature:
    a = np.ones(1500)
    a[75:100] = 5.0
    a[115:125] = 5.0
    a[575:600] = 5.0
    desp = dsp.despike(a, 35, exclude_point="last")
    assert desp.niter < 5

    pv = np.zeros(1500)
    pv[75:100] = True
    pv[115:125] = True
    pv[575:600] = True
    assert (desp.pv == pv).all()
    assert desp.niter < 5


def test_despike_diff():
    x = np.arange(100.0)
    x[45] = 4.5
    x[0] = -20
    x[-13] = 115
    s = dsp.despike_diff(x, 9, sigma=4)
    pv = np.zeros(x.shape, bool)
    pv[0] = True
    pv[45] = True
    pv[-13] = True
    assert (s.pv == pv).all()
    assert (s.x == x[~pv]).all()

    x[45] = 4.5
    x[0] = -20
    x[-13] = 115
    s = dsp.despike_diff(x, 9, sigma=4, maxiter=1)
    pv = np.zeros(x.shape, bool)
    pv[-13] = True
    assert (s.pv == pv).all()
    assert (s.x == x[~pv]).all()
    assert s.niter == 1

    with pytest.raises(ValueError):
        dsp.despike_diff(np.ones((5, 5)), 3)


def test_fftcoef():
    for n in (50, 51):
        t = np.arange(n) / n

        x = 4.0 * np.sin(2 * np.pi * 4 * t)
        mag, phase, frq = dsp.fftcoef(x, 1 / t[1])
        # lowest frequency = 1/T = 1.0
        # highest frequency = sr/2 = 500.0
        nfrq = (n // 2) + 1
        assert np.allclose(frq, 0.0 + np.arange(nfrq))
        x2 = np.zeros(len(frq))
        x2[4] = 4.0
        assert np.allclose(mag, x2)
        assert np.allclose(phase[4], 0.0)

        x = 4.0 * np.cos(2 * np.pi * 4 * t)
        mag, phase, frq = dsp.fftcoef(x, 1 / t[1])
        # lowest frequency = 1/T = 1.0
        # highest frequency = sr/2 = 500.0
        nfrq = (n // 2) + 1
        assert np.allclose(frq, 0.0 + np.arange(nfrq))
        x2 = np.zeros(len(frq))
        x2[4] = 4.0
        assert np.allclose(mag, x2)
        assert np.allclose(phase[4], -np.pi / 2)

        sig = np.column_stack((x, x, x))
        mag1, phase1, frq1 = dsp.fftcoef(sig, 1 / t[1], axis=0)
        assert np.allclose(frq, frq1)
        for i in range(3):
            assert np.allclose(mag1[:, i], mag)
            assert np.allclose(phase1[:, i], phase)

        sig = sig.reshape(1, 1, -1, 3, 1)
        mag2, phase2, frq2 = dsp.fftcoef(sig, 1 / t[1], axis=2)
        assert np.allclose(frq, frq2)
        assert np.allclose(mag1, mag2.reshape(-1, 3))
        assert np.allclose(phase1, phase2.reshape(-1, 3))


def test_fftcoef_recon():
    for n in (50, 51):
        t = np.arange(n) / n
        x = np.random.randn(n)
        mag, phase, frq = dsp.fftcoef(x, 1 / t[1])
        A, B, frq = dsp.fftcoef(x, 1 / t[1], coef="ab", window=np.ones(x.shape))

        w = 2 * np.pi * frq[1]
        # reconstruct with magnitude and phase:
        x2 = 0.0
        for k, (m, p, f) in enumerate(zip(mag, phase, frq)):
            x2 = x2 + m * np.sin(k * w * t - p)
        assert np.allclose(x, x2)

        # reconstruct with A and B:
        x3 = 0.0
        for k, (a, b, f) in enumerate(zip(A, B, frq)):
            x3 = x3 + a * np.cos(k * w * t) + b * np.sin(k * w * t)
        assert np.allclose(x, x3)


def test_fftcoef_fold():
    for n in (10, 11):
        t = np.arange(n) / n
        x = np.random.randn(n)
        mag, phase, frq = dsp.fftcoef(x, 1 / t[1], fold=True)
        mag1, phase1, frq = dsp.fftcoef(x, 1 / t[1], fold=False)
        if not (n & 1):
            mag1[1:-1] *= 2.0
        else:
            mag1[1:] *= 2.0
        assert np.allclose(mag, mag1)
        assert np.allclose(phase, phase1)


def test_fftcoef_detrend():
    for n in (50, 51):
        t = np.arange(n) / n
        x = np.random.randn(n)
        xd = signal.detrend(x)
        mag, phase, frq = dsp.fftcoef(x, 1 / t[1], dodetrend=True)
        A, B, frq = dsp.fftcoef(
            x, 1 / t[1], coef="ab", window=np.ones(x.shape), dodetrend=True
        )

        w = 2 * np.pi * frq[1]
        # reconstruct with magnitude and phase:
        x2 = 0.0
        for k, (m, p, f) in enumerate(zip(mag, phase, frq)):
            x2 = x2 + m * np.sin(k * w * t - p)
        assert np.allclose(xd, x2)

        # reconstruct with A and B:
        x3 = 0.0
        for k, (a, b, f) in enumerate(zip(A, B, frq)):
            x3 = x3 + a * np.cos(k * w * t) + b * np.sin(k * w * t)
        assert np.allclose(xd, x3)

    with pytest.raises(ValueError):
        dsp.fftcoef(x, 1 / t[1], window=np.ones(x.size - 1))


def test_fftcoef_maxdf():
    n = 16
    t = np.arange(n) / n
    x = np.random.randn(n)
    mag, phase, frq = dsp.fftcoef(x, 1 / t[1])
    mag1, phase1, frq1 = dsp.fftcoef(x, 1 / t[1], maxdf=0.6)
    assert frq1[1] <= 0.6
    assert np.allclose(mag1[::2], mag)



from argparse import ArgumentParser

from pyyeti.nastran import op2


def main() -> None:
    """Main."""
    parser = ArgumentParser(description="List contents of op2 file(s)")

    parser.add_argument(
        "input_file", metavar="INFILE", help="file to read", type=str, nargs="+"
    )

    parser.add_argument(
        "-w",
        "--with-headers",
        dest="with_headers",
        action="store_true",
        default=False,
        help="include print of table record headers",
    )

    parser.add_argument(
        "-t",
        "--with-trailers",
        dest="with_trailers",
        action="store_true",
        default=False,
        help="include data block trailers in print out",
    )

    args = parser.parse_args()

    for name in args.input_file:
        with op2.OP2(name) as o2:
            print()
            print(f"{name}: {len(o2.dblist)} data blocks")
            o2.directory(
                with_headers=args.with_headers, with_trailers=args.with_trailers
            )


if __name__ == "__main__":
    main()



import numpy as np
from numpy.random import Generator, MT19937
import tempfile
import os
from pyyeti import ytools
from scipy import linalg
from scipy import interpolate as sint
import pytest


def test_fit_circle_3d():
    parms = ytools.fit_circle_3d(3 * np.eye(3), makeplot="new")
    center = np.array([1.0, 1.0, 1.0])
    radius = np.linalg.norm(np.array([3.0, 0, 0]) - center)
    z_direction = ytools._norm_vec(center)

    assert np.allclose(parms.center, center)
    assert np.allclose(parms.radius, radius)
    assert np.allclose(parms.basic2local[2], z_direction)


def test_histogram():
    assert np.allclose(
        ytools.histogram([1, 1, 3, 3, 4], 1),
        np.array([[1.0, 2.0, 40.0], [3.0, 2.0, 40.0], [4.0, 1.0, 20.0]]),
    )
    assert np.allclose(ytools.histogram([np.inf], 1), np.array([[0.0, 0.0, 0.0]]))


def test_sturm():
    a = np.array([0.0, 0.16, 1.55, 2.78, 9.0, 14.0])
    A = np.diag(a)
    assert np.all(1 == ytools.sturm(A, 0.0))
    assert np.all(3 == ytools.sturm(A, 1.55))


def test_eig_si():
    k = np.random.randn(40, 40)
    m = np.random.randn(40, 40)
    k = np.dot(k.T, k) * 1000
    m = np.dot(m.T, m) * 10
    w1, phi1 = linalg.eigh(k, m, subset_by_index=(0, 14))
    w2, phi2, phiv2 = ytools.eig_si(
        k, m, p=15, mu=-1, tol=1e-12, verbose=False, rng=Generator(MT19937(1))
    )
    w2b, phi2b, phiv2b = ytools.eig_si(
        k, m, p=15, mu=-1, tol=1e-12, verbose=False, rng=Generator(MT19937(1))
    )

    assert (w2b == w2).all()
    assert (phi2b == phi2).all()
    assert (phiv2b == phiv2).all()

    fcut = np.sqrt(w2.max()) / 2 / np.pi * 1.001
    w3, phi3, phiv3 = ytools.eig_si(k, m, f=fcut, mu=-1, tol=1e-12)
    assert np.allclose(w1, w2)
    assert np.allclose(np.abs(phi1), np.abs(phi2))
    assert np.allclose(w1, w3)
    assert np.allclose(np.abs(phi1), np.abs(phi3))
    w4, phi4, phiv4 = ytools.eig_si(k, m, f=fcut, Xk=phiv3, tol=1e-12, pmax=10)
    assert np.allclose(w1[:10], w4)
    assert np.allclose(np.abs(phi1[:, :10]), np.abs(phi4))

    w5, phi5, phiv5 = ytools.eig_si(k, m, p=15, Xk=phi4, tol=1e-12)
    assert np.allclose(w1, w5)
    assert np.allclose(np.abs(phi1), np.abs(phi5))

    mmod = m.copy()
    mmod[1, 0] = 2 * mmod[1, 0]
    with pytest.raises(ValueError):
        ytools.eig_si(k, mmod, p=15, Xk=phi4, tol=1e-12)


def test_gensweep():
    sig, t, f = ytools.gensweep(10, 1, 12, 8)
    assert np.allclose(np.max(sig), 1.0)
    assert np.allclose(np.min(sig), -1.0)
    # 12 = 1*2**n_oct
    n_oct = np.log(12.0) / np.log(2.0)
    t_elapsed = n_oct / 8 * 60
    assert np.abs(t[-1] - t_elapsed) < 0.01
    assert np.allclose(f[0], 1)
    assert np.abs(f[-1] - 12) < 0.01
    assert np.allclose(t[1] - t[0], 1 / 10 / 12)
    # at 8 oct/min or 8/60 = 4/30 = 2/15 oct/sec, time for 3 octaves
    # is: 3/(2/15) = 45/2 = 22.5 s
    i = np.argmin(np.abs(t - 22.5))
    assert np.allclose(f[i], 8.0)


def test_mattype():
    t, m = ytools.mattype([1, 2, 3])
    assert t == 0
    assert not ytools.mattype([1, 2, 3], "symmetric")

    a = np.random.randn(4, 4)
    a = a.dot(a.T)
    t, m = ytools.mattype(a)
    assert t & m["posdef"] and t & m["symmetric"]
    assert ytools.mattype(a, "posdef")
    assert ytools.mattype(a, "symmetric")

    a[1, 1] = 0
    t, m = ytools.mattype(a)
    assert (not (t & m["posdef"])) and t & m["symmetric"]
    assert not ytools.mattype(a, "posdef")
    assert ytools.mattype(a, "symmetric")

    c = np.random.randn(4, 4) + 1j * np.random.randn(4, 4)
    c = c.dot(np.conj(c.T))
    t, m = ytools.mattype(c)
    assert t & m["posdef"] and t & m["hermitian"]
    assert ytools.mattype(c, "posdef")
    assert ytools.mattype(c, "hermitian")

    c[1, 1] = 0
    t, m = ytools.mattype(c)
    assert (not (t & m["posdef"])) and t & m["hermitian"]
    assert ytools.mattype(c, "hermitian")
    assert not ytools.mattype(c, "posdef")

    assert ytools.mattype(np.eye(5), "diagonal")
    assert ytools.mattype(np.eye(5), "identity")
    assert not ytools.mattype(np.eye(5) * 2.0, "identity")
    assert not ytools.mattype(np.random.randn(5, 5), "diagonal")
    assert not ytools.mattype(np.random.randn(5, 5), "identity")
    assert not ytools.mattype(np.random.randn(5, 5), "posdef")
    assert not ytools.mattype(np.random.randn(5, 5) * (2 + 1j), "posdef")

    with pytest.raises(ValueError):
        ytools.mattype(c, "badtype")


def test_mattype2():
    # real a
    a = np.array([[346500.0, 1e-7], [2.1e-7, 1000000.1]])
    assert 13 == ytools.mattype(a)[0]
    t, mattypes, ch = ytools.mattype(a, return_cholesky=True)
    assert t == 13
    assert np.allclose(ch, linalg.cholesky(a))

    # complex a
    a = np.array([[346500.0, 1e-7 * (1 + 1j)], [2.1e-7 * (1 - 1j), 1000000.1]])
    assert 15 == ytools.mattype(a)[0]
    t, mattypes, ch = ytools.mattype(a, return_cholesky=True)
    assert t == 15
    assert np.allclose(ch, linalg.cholesky(a))

    # real a
    a = np.array([[346500.0, 1e-7], [2000.1e-7, 1000000.1]])
    assert 0 == ytools.mattype(a)[0]
    t, mattypes, ch = ytools.mattype(a, return_cholesky=True)
    assert t == 0
    assert ch is None

    # complex a
    a = np.array([[346500.0, 1e-7 * (1 + 1j)], [2000.1e-7 * (1 - 1j), 1000000.1]])
    assert 0 == ytools.mattype(a)[0]
    t, mattypes, ch = ytools.mattype(a, return_cholesky=True)
    assert t == 0
    assert ch is None


def test_save_load():
    a = np.arange(18).reshape(2, 3, 3)
    b = np.arange(3)
    d = dict(A="test var", B=[1, 2, 3])

    names = []
    try:
        f = tempfile.NamedTemporaryFile(delete=False)
        name = f.name
        f.close()
        obj = dict(a=a, b=b, c=a, d=d, e=d)
        for ext in ("", ".pgz", ".pbz2"):
            fname = name + ext
            names.append(fname)
            ytools.save(fname, obj)
            obj_in = ytools.load(fname)
            assert np.all(obj["a"] == obj_in["a"])
            assert np.all(obj["b"] == obj_in["b"])
            assert obj_in["c"] is obj_in["a"]
            assert obj["d"] == obj_in["d"]
            assert obj_in["e"] is obj_in["d"]
    finally:
        for name in names:
            os.remove(name)


def test_compmat():
    A = np.array([[1, 4, 5, 40], [15, 16, 17, 80]])
    B = np.array([[2, 4.2, 5, 43], [20, 14, 17, 82]])
    a, s = ytools.compmat(A, B, 0.1, method="row", pdiff_tol=5, verbose=0)
    assert np.allclose(a, 33.333333333333)

    A = A + B * 1j
    B = B + A.real * 1j
    a, s = ytools.compmat(A, B, 0.04, method="row", pdiff_tol=5, verbose=0)

    # A = array([[ 1. +2.j ,  4. +4.2j,  5. +5.j , 40.+43.j ],
    #            [15.+20.j , 16.+14.j , 17.+17.j , 80.+82.j ]])
    #
    # B = array([[ 2.  +1.j,  4.2 +4.j,  5.  +5.j, 43. +40.j],
    #            [20. +15.j, 14. +16.j, 17. +17.j, 82. +80.j]])

    real_pdiff = (B.real / A.real - 1) * 100
    # array([[100.        ,   5.        ,   0.        ,   7.5       ],
    #        [ 33.33333333, -12.5       ,   0.        ,   2.5       ]])

    imag_pdiff = (B.imag / A.imag - 1) * 100
    # array([[-50.        ,  -4.76190476,   0.        ,  -6.97674419],
    #        [-25.        ,  14.28571429,   0.        ,  -2.43902439]])

    assert a == 100
    real_sbe = np.array(
        [
            real_pdiff.min(),
            real_pdiff.max(),
            real_pdiff.mean(),
            real_pdiff.std(ddof=1),
        ]
    )
    imag_sbe = np.array(
        [
            imag_pdiff.min(),
            imag_pdiff.max(),
            imag_pdiff.mean(),
            imag_pdiff.std(ddof=1),
        ]
    )

    assert np.allclose(s, np.vstack((real_sbe, imag_sbe)))

    a, s = ytools.compmat(A, B.real, 0.04, method="row", pdiff_tol=5, verbose=0)
    assert np.allclose(s, np.vstack((real_sbe, [-100.0, -100.0, -100.0, 0.0])))

    a, s = ytools.compmat(
        A.real, B.real + 0j, 0.04, method="row", pdiff_tol=5, verbose=0
    )
    assert a == 100.0

    a, s = ytools.compmat(A, B, 0.2, method="col", pdiff_tol=5, verbose=0)
    assert np.allclose(a, 33.333333333333)

    a, s = ytools.compmat(A, B, 0.5, method="max", pdiff_tol=5, verbose=0)
    assert np.allclose(a, 7.5)

    a, s = ytools.compmat(A, B, 40.0, method="abs", pdiff_tol=5, verbose=0)
    assert np.allclose(a, 7.5)

    with pytest.raises(ValueError):
        ytools.compmat(A, A[:2, :2])

    with pytest.raises(ValueError):
        ytools.compmat(A, B, method="badmethod")

    a = np.zeros((3, 3, 3))
    with pytest.raises(ValueError):
        ytools.compmat(a, a)


def test_max_complex_vector_sum():
    x = [3.0, 1.0 + 2.0j]
    y = [4.0, 3.0 + 4.0j]
    h, th, c, s = ytools.max_complex_vector_sum(x, y)

    # check for consistent output:
    for xi, yi, hi, thi, ci, si in zip(x, y, h, th, c, s):
        assert np.allclose(xi * ci + yi * si, hi)
        assert np.allclose(np.cos(thi), ci)
        assert np.allclose(np.sin(thi), si)

    # check for correct solution:
    mx = np.zeros(len(x))
    for i, (xi, yi) in enumerate(zip(x, y)):
        for a in np.arange(0.0, np.pi, 0.001):
            mx[i] = max(mx[i], np.abs(np.cos(a) * xi + np.sin(a) * yi))
    assert (np.abs(h) >= mx).all()

try:
    import numba
except ImportError:
    pass
else:

    def test_numba_interp_32():
        fp = np.array([[2, 1, 4]])
        fp = np.tile(fp, (100, 1))
        xp = np.array([1., 2., 3.])
        x = np.arange(0, 4, 0.1)

        f = ytools.numba_interp(x, xp, fp.astype(np.float32))

        assert f.dtype == np.float32

        # Compare against SciPy Interpolate, interp1d - with "hold value" extrapolation instead of linear extrapolation
        interp_func = sint.interp1d(xp, fp, bounds_error=False, fill_value=(fp[:, 0], fp[:, -1]))
        f_scipy = interp_func(x)

        np.testing.assert_array_almost_equal(f, f_scipy, decimal=7)


    def test_numba_interp_64():
        fp = np.array([[2, 1, 4]])
        fp = np.tile(fp, (100, 1))
        xp = np.array([1., 2., 3.])
        x = np.arange(0, 4, 0.1)

        f = ytools.numba_interp(x, xp, fp.astype(np.int32))
        assert f.dtype == np.float64

        f = ytools.numba_interp(x, xp, fp.astype(np.int64))
        assert f.dtype == np.float64

        f = ytools.numba_interp(x, xp, fp.astype(np.float64))
        assert f.dtype == np.float64

        # Compare against SciPy Interpolate, interp1d - with "hold value" extrapolation instead of linear extrapolation
        interp_func = sint.interp1d(xp, fp, bounds_error=False, fill_value=(fp[:, 0], fp[:, -1]))
        f_scipy = interp_func(x)

        np.testing.assert_array_equal(f, f_scipy)



import re
import numpy as np
import scipy.linalg as la
from pyyeti import era, ode
import pytest
import warnings


def get_model_data():
    M = np.identity(3)
    K = np.array(
        [
            [4185.1498, 576.6947, 3646.8923],
            [576.6947, 2104.9252, -28.0450],
            [3646.8923, -28.0450, 3451.5583],
        ]
    )
    D = np.array(
        [
            [4.96765646, 0.97182432, 4.0162425],
            [0.97182432, 6.71403672, -0.86138453],
            [4.0162425, -0.86138453, 4.28850828],
        ]
    )

    (w2, phi) = la.eigh(K, M)
    omega = np.sqrt(w2)
    freq_hz = omega / (2 * np.pi)
    modal_damping = phi.T @ D @ phi
    zeta = np.diag(modal_damping) / (2 * omega)

    return M, K, D, freq_hz, zeta, phi


def test_era():
    M, K, D, freq_hz, zeta, phi = get_model_data()

    dt = 0.01
    t = np.arange(0, 1, dt)
    F = np.zeros((3, len(t)))
    ts = ode.SolveExp2(M, D, K, dt)
    sol = ts.tsolve(force=F, v0=[1, 1, 1])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        era_fit = era.ERA(
            sol.v,
            sr=1 / dt,
            auto=True,
            input_labels=["x", "y", "z"],
            all_lower_limits=0.1,
            FFT=True,
            # FFT_range=(0.0, 25.0),
            # show_plot=False,
            verbose=False,
        )

    assert np.allclose(era_fit.freqs_hz, freq_hz)
    assert np.allclose(era_fit.zeta, zeta)
    phi_rat = phi / era_fit.phi
    phi_rat = phi_rat / phi_rat[0]
    assert np.allclose(np.ones((3, 3)), phi_rat)

    for val in era_fit.lower_limits.values():
        assert np.isclose(0.1, val)

    # different size H:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        era_fit = era.ERA(
            sol.v,
            sr=1 / dt,
            auto=True,
            alpha=65,
            input_labels=["x", "y", "z"],
            all_lower_limits=0.1,
            FFT=True,
            FFT_range=(0.0, 25.0),
            # show_plot=False,
            verbose=False,
        )

    assert np.allclose(era_fit.freqs_hz, freq_hz)
    assert np.allclose(era_fit.zeta, zeta)
    phi_rat = phi / era_fit.phi
    phi_rat = phi_rat / phi_rat[0]
    assert np.allclose(np.ones((3, 3)), phi_rat)

    # different size H 2:
    era_fit = era.ERA(
        sol.v,
        sr=1 / dt,
        auto=True,
        beta=65,
        input_labels=["x", "y", "z"],
        all_lower_limits=0.1,
        show_plot=False,
        verbose=False,
    )

    assert np.allclose(era_fit.freqs_hz, freq_hz)
    assert np.allclose(era_fit.zeta, zeta)
    phi_rat = phi / era_fit.phi
    phi_rat = phi_rat / phi_rat[0]
    assert np.allclose(np.ones((3, 3)), phi_rat)

    # different size H 3:
    era_fit = era.ERA(
        sol.v,
        sr=1 / dt,
        auto=True,
        alpha=55,
        beta=40,
        input_labels=["x", "y", "z"],
        all_lower_limits=0.1,
        show_plot=False,
        verbose=False,
    )

    assert np.allclose(era_fit.freqs_hz, freq_hz)
    assert np.allclose(era_fit.zeta, zeta)
    phi_rat = phi / era_fit.phi
    phi_rat = phi_rat / phi_rat[0]
    assert np.allclose(np.ones((3, 3)), phi_rat)

    # different usage of svd_tol:
    era_fit = era.ERA(
        sol.v,
        sr=1 / dt,
        auto=True,
        svd_tol=6,
        input_labels=["x", "y", "z"],
        show_plot=False,
        verbose=False,
    )

    assert np.allclose(era_fit.freqs_hz, freq_hz)
    assert np.allclose(era_fit.zeta, zeta)
    phi_rat = phi / era_fit.phi
    phi_rat = phi_rat / phi_rat[0]
    assert np.allclose(np.ones((3, 3)), phi_rat)

    with pytest.raises(RuntimeError, match="No modes selected"):
        era_fit = era.ERA(
            sol.v,
            sr=1 / dt,
            auto=True,
            svd_tol=6,
            input_labels=["x", "y", "z"],
            all_lower_limits=1.1,  # force failure
            show_plot=False,
            verbose=False,
        )


def test_NExT():
    with pytest.raises(ValueError, match="invalid value for `domain`"):
        era.NExT([1, 2, 3], 100, lag_stop=5, domain="bad value")

    M, K, D, freq_hz, zeta, phi = get_model_data()

    dt = 0.01
    t = np.arange(0, 100, dt)
    sr = 1 / dt
    np.random.seed(1)  # for repeatability
    F = np.random.randn(3, len(t))
    ts = ode.SolveUnc(M, D, K, dt)
    sol = ts.tsolve(force=F)

    era_fit1 = era.ERA(
        era.NExT(sol.a, sr, lag_stop=75),
        sr=sr,
        auto=True,
        show_plot=False,
        verbose=False,
    )

    # check for equivalence:
    era_fit2 = era.ERA(
        era.NExT(
            sol.a,
            sr,
            lag_stop=75,
            domain="frequency",
            nperseg=10_000,
            window="boxcar",
        ),
        sr=1 / dt,
        auto=True,
        show_plot=False,
        verbose=False,
    )
    assert np.allclose(era_fit1.freqs_hz, era_fit2.freqs_hz)

    # show that unbiased is doing something:
    era_fit2 = era.ERA(
        era.NExT(sol.a, sr, lag_stop=75, unbiased=False),
        sr=1 / dt,
        auto=True,
        show_plot=False,
        verbose=False,
    )
    assert not np.allclose(era_fit1.freqs_hz, era_fit2.freqs_hz)

    # show that demeaning is working:
    era_fit2 = era.ERA(
        era.NExT(sol.a + 10, sr, lag_stop=75),
        sr=1 / dt,
        auto=True,
        show_plot=False,
        verbose=False,
    )
    assert np.allclose(era_fit1.freqs_hz, era_fit2.freqs_hz)

    era_fit2 = era.ERA(
        era.NExT(sol.a + 0.1, sr, lag_stop=75, demean=False),
        sr=1 / dt,
        auto=True,
        show_plot=False,
        verbose=False,
    )
    assert not np.allclose(era_fit1.freqs_hz, era_fit2.freqs_hz)



import numpy as np
from pyyeti import locate
import pytest


def test_find_rows():
    a = [[1, 2, 3], [4, 5, 6]]
    b = [1, 2, 3, 4]
    assert locate.find_rows(a, b).size == 0


def test_mat_intersect():
    a, b = locate.mat_intersect([1, 2, 3, 4, 5], [4, 2, 8])
    assert np.all(a == np.array([3, 1]))
    assert np.all(b == np.array([0, 1]))

    a, b = locate.mat_intersect([1, 2, 3, 4, 5], [4, 2, 8], keep=1)
    assert np.all(a == np.array([1, 3]))
    assert np.all(b == np.array([1, 0]))

    a, b = locate.mat_intersect([1, 2, 3, 4, 5], [4, 2, 8], keep=2)
    assert np.all(a == np.array([3, 1]))
    assert np.all(b == np.array([0, 1]))

    a, b = locate.mat_intersect([1, 2, 3], [1, 2, 3, 1], keep=2)
    assert np.all(a == [0, 1, 2, 0])
    assert np.all(b == [0, 1, 2, 3])

    a, b = locate.mat_intersect([1, 2, 3], [1, 2, 3, 1], keep=1)
    assert np.all(a == [0, 1, 2])
    assert np.all(b == [0, 1, 2])

    a, b = locate.mat_intersect([1, 2, 3], [1, 2, 3, 1])
    assert np.all(a == [0, 1, 2])
    assert np.all(b == [0, 1, 2])

    a, b = locate.mat_intersect([1, 2, 3], [100, 200, 300, 100])
    assert len(a) == len(b) == 0

    # now with floating point:
    a, b = locate.mat_intersect([1.0, 2, 3, 4, 5], [4, 2, 8.0])
    assert np.all(a == np.array([3, 1]))
    assert np.all(b == np.array([0, 1]))

    a, b = locate.mat_intersect([1, 2, 3, 4, 5], [4, 2, 8.0], keep=1)
    assert np.all(a == np.array([1, 3]))
    assert np.all(b == np.array([1, 0]))

    a, b = locate.mat_intersect([1, 2, 3, 4, 5], [4, 2.0, 8], keep=2)
    assert np.all(a == np.array([3, 1]))
    assert np.all(b == np.array([0, 1]))

    a, b = locate.mat_intersect([1, 2, 3], [1, 2, 3, 1.0], keep=2)
    assert np.all(a == [0, 1, 2, 0])
    assert np.all(b == [0, 1, 2, 3])

    a, b = locate.mat_intersect([1, 2, 3], [1, 2, 3, 1.0], keep=1)
    assert np.all(a == [0, 1, 2])
    assert np.all(b == [0, 1, 2])

    a, b = locate.mat_intersect([1, 2, 3], [1, 2, 3, 1.0])
    assert np.all(a == [0, 1, 2])
    assert np.all(b == [0, 1, 2])

    a, b = locate.mat_intersect([1, 2, 3], [100, 200, 300, 100.0])
    assert len(a) == len(b) == 0


def test_mat_intersect_2():
    n = 30_000
    haystack = np.random.randn(n, 33)
    needles = haystack[::7]
    pv1, pv2 = locate.mat_intersect(haystack, needles, keep=2)
    assert (pv1 == np.arange(0, n, 7)).all()
    assert (pv2 == np.arange(0, n // 7 + 1)).all()


def test_locate_misc():
    mat1 = np.array([[7, 3], [6, 8], [4, 0], [9, 2], [1, 5]])
    mat2 = np.array([[9, 2], [1, 5], [7, 3]])
    pv1, pv2 = locate.mat_intersect(mat1, mat2)
    assert np.all(np.array([3, 4, 0]) == pv1)
    assert np.all(np.array([0, 1, 2]) == pv2)
    pv1, pv2 = locate.mat_intersect(mat1, mat2, 1)
    assert np.all(np.array([0, 3, 4]) == pv1)
    assert np.all(np.array([2, 0, 1]) == pv2)
    pv1, pv2 = locate.mat_intersect(mat1, mat2, 2)
    assert np.all(np.array([3, 4, 0]) == pv1)
    assert np.all(np.array([0, 1, 2]) == pv2)
    pv = np.array([0, 3, 5])
    tf = locate.index2bool(pv, 8)
    assert np.all(np.array([True, False, False, True, False, True, False, False]) == tf)


def test_index2slice():
    from pyyeti.locate import index2slice

    a = np.arange(50, 100)

    pv = [3, 4, 5, 6]
    s = index2slice(pv)
    assert s == slice(3, 7, 1)
    assert (a[pv] == a[s]).all()

    pv = [10, 7, 4]
    s = index2slice(pv)
    assert s == slice(10, 1, -3)
    assert (a[pv] == a[s]).all()

    pv = [-1]
    s = index2slice(pv)
    assert s == slice(-1, None, None)
    assert (a[pv] == a[s]).all()

    pv = [-3]
    s = index2slice(pv)
    assert s == slice(-3, -2, None)
    assert (a[pv] == a[s]).all()

    pv = np.array([1, 6, 7])
    s = index2slice(pv)
    assert s is pv

    with pytest.raises(ValueError, match="invalid partition vector"):
        index2slice(pv, strict=True)

    with pytest.raises(ValueError, match="`pv` has 2 dimensions"):
        index2slice([[3]])



# Simulate event and recover responses
import numpy as np
from scipy.io import matlab
import matplotlib as mpl

mpl.interactive(0)
mpl.use("Agg")

from pyyeti import n2p, op2, stats, ode, cla
from pyyeti.pp import PP

# event name:
event = "TOECO"

if 1:
    # load data recovery data:
    sc = cla.load("../cla_params.pgz")
    cla.PrintCLAInfo(sc["mission"], event)

    # load nastran data:
    nas = op2.rdnas2cam("../nas2cam")

    # form ulvs for some SEs:
    SC = 101
    n2p.addulvs(nas, SC)

    import sys

    sys.path.insert(0, "..")
    import alphajoint

    drdefs = alphajoint.get_drdefs(nas, sc)

    # prepare spacecraft data recovery matrices
    DR = cla.DR_Event()
    DR.add(nas, sc["drdefs"])
    DR.add(nas, drdefs)

    # initialize results (ext, mnc, mxc for all drms)
    results = DR.prepare_results(sc["mission"], event)

    # set rfmodes:
    rfmodes = nas["rfmodes"][0]

    # setup modal mass, damping and stiffness
    m = None  # None means identity
    k = nas["lambda"][0]
    k[: nas["nrb"]] = 0.0
    b = 2 * 0.02 * np.sqrt(k)
    mbk = (m, b, k)

    # load in forcing functions:
    toeco = matlab.loadmat("toeco_ffns.mat", squeeze_me=True, struct_as_record=False)
    toeco["ffns"] = toeco["ffns"][:2, ::2]
    toeco["sr"] = toeco["sr"] / 2
    toeco["t"] = toeco["t"][::2]

    # form force transform:
    T = n2p.formdrm(nas, 0, [[8, 12], [24, 13]])[0].T

    # do pre-calcs and loop over all cases:
    ts = ode.SolveUnc(*mbk, 1 / toeco["sr"], rf=rfmodes)
    LC = toeco["ffns"].shape[0]
    t = toeco["t"]
    for j, force in enumerate(toeco["ffns"]):
        print("Running {} case {}".format(event, j + 1))
        genforce = T @ ([[1], [0.1], [1], [0.1]] * force[None, :])
        # solve equations of motion
        sol = ts.tsolve(genforce, static_ic=1)
        sol.t = t
        sol = DR.apply_uf(sol, *mbk, nas["nrb"], rfmodes)
        caseid = "{} {:2d}".format(event, j + 1)
        results.time_data_recovery(sol, nas["nrb"], caseid, DR, LC, j)

    # save results:
    cla.save("results.pgz", results)

#     # make some srs plots and tab files:
#     # results.rptext()
#     # results.rpttab()
#     results.rpttab(excel='toeco')
#     results.srs_plots()
#     # results.srs_plots(fmt='png')
#     results.resp_plots()



from pyyeti import cla
from pyyeti.nastran import n2p


def alphajoint(sol, nas, Vars, se):
    return Vars[se]["alphadrm"] @ sol.a


def get_drdefs(nas, sc):
    drdefs = cla.DR_Def(sc["drdefs"].defaults)

    @cla.DR_Def.addcat
    def _():
        se = 0
        name = "alphajoint"
        desc = "Alpha-Joint Acceleration"
        units = "mm/sec^2, rad/sec^2"
        labels = ["Alpha-Joint {:2s}".format(i) for i in "X,Y,Z,RX,RY,RZ".split(",")]
        drms = {"alphadrm": n2p.formdrm(nas, 0, 33)[0]}
        srsopts = dict(eqsine=1, ic="steady")
        histpv = 1  # second row
        srspv = [1]
        drdefs.add(**locals())

    return drdefs
